/* Copyright (C) 2006 Jacco de Leeuw <jacco2@dds.nl>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 */

// pfximprt1Dlg.cpp : implementation file
//

#include "stdafx.h"
#include "pfximprt1.h"
#include "pfximprt1Dlg.h"

#include <wincrypt.h>

#pragma comment(lib, "crypt32.lib")
#include <shlobj.h>
#pragma comment(lib, "ceshell.lib")

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

void _err(const wchar_t* fcn, DWORD err = GetLastError()) 
{
	wchar_t szErrMsg[256]=_T("");
	wchar_t szMsg[512]=_T("");
	if ( FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM, 0, err, 0, szErrMsg, sizeof szErrMsg / sizeof *szErrMsg, 0 ) )
	{
		 swprintf( szMsg, L"%s failed: %s", fcn, szErrMsg );
	}else 
	{
		swprintf( szMsg, L"%s failed: 0x%08X", fcn, err );
	}
	AfxMessageBox(szMsg, MB_ICONSTOP);
}

// Determine if the certificate has the "CA:TRUE" X.509 basic constraint,
// i.e. if it is a Root/CA certificate.
//
BOOL CertIsCACert(PCCERT_CONTEXT pCtx) 
{
	PCERT_EXTENSION pCertExt = NULL;
	BOOL fCA = FALSE; 
	PCERT_BASIC_CONSTRAINTS2_INFO pInfo = NULL;
	DWORD cbInfo = 0;

	pCertExt = CertFindExtension(	szOID_BASIC_CONSTRAINTS2,
									pCtx->pCertInfo->cExtension,
									pCtx->pCertInfo->rgExtension
								);
	if (pCertExt == NULL) 
	{
		// If no extension found, we assume it is a personal cert
		return FALSE;
	}

	if (!CryptDecodeObjectEx(	X509_ASN_ENCODING,
								X509_BASIC_CONSTRAINTS2,
								pCertExt->Value.pbData,
								pCertExt->Value.cbData,
								CRYPT_DECODE_ALLOC_FLAG,
								(PCRYPT_DECODE_PARA)NULL,
								&pInfo,
								&cbInfo))
	{
		// Cannot decode certificate extension.
		// XXX Then we decide it is a personal personal cert
		// instead of exiting the program with _err(L"Cannot decode");
		return FALSE;
	} 

	if (pInfo)
	{
		fCA = pInfo->fCA;
		LocalFree(pInfo);
	}
	return fCA; 
}

// Determine if the certificate has is self-signed,
// i.e. if the subject equals the issuer. Then it is a Root certificate.
//
BOOL CACertIsSelfsigned(PCCERT_CONTEXT pCtx)
{
	DWORD dwFlags = CERT_STORE_SIGNATURE_FLAG;
	
	if (!(CertCompareCertificateName(	X509_ASN_ENCODING,
										&pCtx->pCertInfo->Issuer,
										&pCtx->pCertInfo->Subject)))	
	{
		// wprintf(L"Not self-signed: Issuer != Subject\n");
		return FALSE;
	}

	if  (!(CertVerifySubjectCertificateContext(	pCtx, pCtx, &dwFlags)))
	{
		// wprintf(L"Not self-signed: CertVerifySubjectCertificateContext\n");
        return FALSE;
	}

	if (dwFlags != 0)
	{
		// wprintf(L"Not self-signed: dwflags\n");
		return FALSE;
	}

    return TRUE; 
}

// Cpfximprt1Dlg dialog

Cpfximprt1Dlg::Cpfximprt1Dlg(CWnd* pParent /*=NULL*/)
	: CDialog(Cpfximprt1Dlg::IDD, pParent)
	, m_CertificateFile(_T(""))
	, m_Password(_T(""))
{
	TCHAR cert[MAX_PATH]=_T("");
	
	m_hIcon = AfxGetApp()->LoadIcon(IDR_MAINFRAME);

	// By default, we will look for the default PKCS#12 file
	// in the "My Documents" folder.
	//
	// Find "My Documents" folder (actual name depends on PPC language version).
	if (!SHGetDocumentsFolder(_T("\\"),cert))
	{
		// Fallback to the English default
		_tcscpy(cert, _T("\\My Documents"));		
	}
	// Append slash and default filename.
	_tcscat(cert, _T("\\user.pfx"));
	// Use this file and path as the default
	m_CertificateFile = cert;

}

void Cpfximprt1Dlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	DDX_Text(pDX, IDC_EDIT1, m_CertificateFile);
	DDX_Text(pDX, IDC_EDIT2, m_Password);
}

BEGIN_MESSAGE_MAP(Cpfximprt1Dlg, CDialog)
	//}}AFX_MSG_MAP
	ON_EN_CHANGE(IDC_EDIT1, &Cpfximprt1Dlg::OnEnChangeEdit1)
	ON_BN_CLICKED(IDC_BUTTON1, &Cpfximprt1Dlg::OnBnClickedButton1)
	ON_BN_CLICKED(IDC_BUTTON3, &Cpfximprt1Dlg::OnBnClickedButton3)
END_MESSAGE_MAP()


// Cpfximprt1Dlg message handlers

BOOL Cpfximprt1Dlg::OnInitDialog()
{
	CDialog::OnInitDialog();
	// Set the icon for this dialog.  The framework does this automatically
	//  when the application's main window is not a dialog
	SetIcon(m_hIcon, TRUE);			// Set big icon
	SetIcon(m_hIcon, FALSE);		// Set small icon

	// TODO: Add extra initialization here

	return TRUE;  // return TRUE  unless you set the focus to a control
}

void Cpfximprt1Dlg::OnEnChangeEdit1()
{
	// TODO:  If this is a RICHEDIT control, the control will not
	// send this notification unless you override the CDialog::OnInitDialog()
	// function and call CRichEditCtrl().SetEventMask()
	// with the ENM_CHANGE flag ORed into the mask.

	// TODO:  Add your control notification handler code here
}

void Cpfximprt1Dlg::OnBnClickedButton1()
{
	// TODO: Add your control notification handler code here
	UpdateData(true);
	CFileDialog dlg(TRUE,
					NULL,
					NULL,
					OFN_HIDEREADONLY,
					_T("Personal certs (*.p12;*.pfx)|*.p12;*.pfx|All files (*.*)|*.*||"),
					NULL);

	// If the user selected a file, put the filename into the field.
	if(dlg.DoModal() == IDOK)
	{
		m_CertificateFile = dlg.GetPathName();
		UpdateData(false);
	}
}

void Cpfximprt1Dlg::OnBnClickedButton3()
{
	CString szResult;
	HCERTSTORE pfxStore = 0;
	HCERTSTORE myStore = 0;
	HCERTSTORE rootStore = 0;
	HCERTSTORE CAStore = 0;
	HCERTSTORE thisCertStore = 0;
	HCERTSTORE hFileStore = 0;
	HANDLE hsection = 0;
	void* pfx = NULL;
	HANDLE hfile = INVALID_HANDLE_VALUE;
	PCCERT_CONTEXT pctx = NULL;
	bool exportable = false;
	unsigned long counter = 0;
	int nChoice = 0;
	wchar_t name[128] = _T("");
	wchar_t szMsg[256] = _T("");
	wchar_t certType[128]=_T("");
	// PKCS#12 file may contain multiple certificates.
	// These vars keep count of the numbers that were imported.
	int nImportedMYcerts = 0;
	int nImportedRootcerts = 0;
	int nImportedCAcerts = 0;
	// Keep count of the type of certificate that was added in an iteration.
	int nRoottoimport = 0;
	int nCAtoimport = 0;
	int nMYtoimport = 0;
	
	// Get the password and the filename from the dialog
	UpdateData(true);

	hfile = CreateFile (m_CertificateFile,   // Open MYFILE.TXT
                     GENERIC_READ,           // Open for reading
                     FILE_SHARE_READ,        // Share for reading
                     NULL,                   // No security
                     OPEN_EXISTING,          // Existing file only
                     FILE_ATTRIBUTE_NORMAL,  // Normal file
                     NULL);                  // No template file

	if (INVALID_HANDLE_VALUE == hfile) 
	{
		_err(L"OpenFile");
		goto cleanup;
	}

	hsection = CreateFileMapping(	hfile, 
									0, 
									PAGE_READONLY, 
									0, 
									0, 
									0);
	if (!hsection) 
	{
		_err(L"CreateFileMapping");
		goto cleanup;
	}

	pfx = MapViewOfFile(	hsection, 
							FILE_MAP_READ, 
							0, 
							0, 
							0);
	if (!pfx) 
	{
		_err(L"MapViewOfFile");
		goto cleanup;
	}

	CRYPT_DATA_BLOB blob;
	blob.cbData = GetFileSize(	hfile, 
								0);
	blob.pbData = (BYTE*)pfx;

	if (!PFXIsPFXBlob(&blob)) 
	{
		AfxMessageBox(L"PFXIsPFXBlob failed: not a valid PKCS#12 file", MB_ICONSTOP );
		goto cleanup;
	}

	//getPassword(password, sizeof password / sizeof *password);
	
	DWORD importFlags = CRYPT_MACHINE_KEYSET;
	//DWORD importFlags = CRYPT_USER_KEYSET;
	if (exportable) 
	{
		importFlags |= CRYPT_EXPORTABLE;
	}
	
	pfxStore = PFXImportCertStore(	&blob, 
									m_Password, 
									importFlags);
	
	// The password is no longer needed so it can be wiped.
	// Let's assume their implementation is secure...
	m_Password.Empty();
	
	if (!pfxStore) 
	{
		_err(L"PFXImportCertStore");
		goto cleanup;
	}

	myStore = CertOpenStore(	CERT_STORE_PROV_SYSTEM, 
								0, 
								0, 
								CERT_SYSTEM_STORE_CURRENT_USER, 
								L"MY");
	if (!myStore) 
	{
		_err(L"CertOpenSystemStore MY");
		goto cleanup;
	}
	rootStore = CertOpenStore(	CERT_STORE_PROV_SYSTEM, 
								0, 
								0, 
								CERT_SYSTEM_STORE_CURRENT_USER, 
								L"Root");
	if (!rootStore ) 
	{
		_err(L"CertOpenSystemStore Root");
		goto cleanup;
	}

	CAStore = CertOpenStore(	CERT_STORE_PROV_SYSTEM, 
								0, 
								0, 
								CERT_SYSTEM_STORE_CURRENT_USER, 
								L"CA");
	if (!CAStore ) 
	{
		_err(L"CertOpenSystemStore CA");
		goto cleanup;
	}
	nImportedMYcerts = 0;
	nImportedRootcerts = 0;
	nImportedCAcerts = 0;
	
	// We assume there are zero or more certificates. If a certificate has the
	// "CA:TRUE" basic contraint (X.509) it will be added to the CA store,
	// unless it is also self-signed, then it will be added to the Root store.
	// All other certificates are considered personal and are added to the MY store.
	while (0 != (pctx = CertEnumCertificatesInStore(pfxStore, pctx))) 
	{
		nRoottoimport = 0;
		nCAtoimport = 0;
		nMYtoimport = 0;		
		
		if (CertGetNameString(pctx, CERT_NAME_FRIENDLY_DISPLAY_TYPE, 0, 0, name, sizeof name / sizeof *name)) 
		{
			// wprintf(L"Found a certificate in the PFX file: %s\n", name);
		}else 
		{
			_err(L"CertGetNameString");
			// XXX Do we exit?
		}
		
		if (CertIsCACert(pctx) == TRUE)
		{
			if (CACertIsSelfsigned(pctx) == TRUE)
			{
				thisCertStore = rootStore;
				nRoottoimport = 1;
				_tcscpy(certType, _T("Root"));				
				// wprintf(L"Attempting to import certificate %d into machine store 'Root'...\n", counter+1);
			} else
			{
				thisCertStore = CAStore;
				nCAtoimport = 1;
				_tcscpy(certType, _T("CA"));
				// wprintf(L"Attempting to import certificate %d into machine store 'CA'...\n", counter+1);			
			}
		} else 
		{
			thisCertStore = myStore;
			nMYtoimport = 1;
			_tcscpy(certType, _T("Personal"));			
			// wprintf(L"Attempting to import certificate %d into machine store 'MY'...\n", counter+1);
		}

		if (CertAddCertificateContextToStore(thisCertStore, pctx, CERT_STORE_ADD_NEW, 0)) 
		{
			// wprintf(L"Import of this certificate succeeded.\n");
			nImportedMYcerts += nMYtoimport;
			nImportedCAcerts += nCAtoimport;
			nImportedRootcerts += nRoottoimport;
		}else 
		{
			DWORD err = GetLastError();
			if (CRYPT_E_EXISTS == err) 
			{
				swprintf( szMsg, L"An equivalent %s cert already exists for '%s'. Overwrite?", certType, name);
				nChoice = AfxMessageBox(szMsg, MB_YESNOCANCEL | MB_ICONQUESTION );
				if (nChoice == IDCANCEL) {
					goto cleanup2;
				}
				if (nChoice == IDYES)
				{
					if (CertAddCertificateContextToStore(thisCertStore, pctx, CERT_STORE_ADD_REPLACE_EXISTING, 0)) 
					{
						// wprintf(L"Import of this certificate succeeded.\n");
						nImportedMYcerts += nMYtoimport;
						nImportedCAcerts += nCAtoimport;
						nImportedRootcerts += nRoottoimport;
					}else 
					{
						_err(L"CertAddCertificateContextToStore");
						goto cleanup2;
					}
				}else 
				{
					// wprintf(L"Skipped.\n");
				}
			}else 
			{
				_err(L"CertAddCertificateContextToStore");
				goto cleanup2;
			}
		}
		++counter;

		}

cleanup2:
	if ((nImportedMYcerts + nImportedRootcerts + nImportedCAcerts) == 0)
	{
		AfxMessageBox(_T("No certificates imported."), MB_ICONINFORMATION);
	} else
	{
		swprintf( szMsg, L"Personal certs: %d\nRoot certs: %d\nIntermed. certs: %d\nTotal imported: %d", nImportedMYcerts, nImportedRootcerts, nImportedCAcerts, nImportedMYcerts + nImportedRootcerts + nImportedCAcerts);
		AfxMessageBox(szMsg, MB_ICONINFORMATION);
	}

cleanup:
	if (myStore) 
	{
		CertCloseStore(myStore, 0);
	}
	if (rootStore) 
	{
		CertCloseStore(rootStore, 0);
	}
	if (CAStore) 
	{
		CertCloseStore(CAStore, 0);
	}
	if (pfxStore) 
	{
		CertCloseStore(pfxStore, CERT_CLOSE_STORE_FORCE_FLAG);
	}
	if (pctx)
		CertFreeCertificateContext(pctx);
	
	if (pfx) 
	{
		UnmapViewOfFile(pfx);
	}
	if (hsection) 
	{
		CloseHandle(hsection);
	}
	if (INVALID_HANDLE_VALUE != hfile) 
	{
		CloseHandle(hfile);
	}
	// Set (wipe) the password in the dialog window
	UpdateData(false);

}
